#include "stock.h"
#include <readline/readline.h>
#include <readline/history.h>
#include <netdb.h>

static pthread_t th_b;

int server, server2;
req_t ok_s='c', err='e', ok_r;
int passwd = 123456;
plid_t plid = -1;
char *prompt_str = "stock > ";

void quit_game();
void login_newplayer ();
void re_entry();
void prompt();			       /* show prompt */
void parse_input(char *s);	       /* Handle input */
void quote (int a, int b, int c, int d);
void ask_comp (cid_t, cid_t);	       /* ask for buy, sell lists */
void recv_comp (cid_t c1, cid_t c2);		       /* get data on companies */
void disp_help ();		       /* show help */
void channel1 ();		       /* stuff through port 5000 */
void *channel2 ();		       /* stuff through 5001 */

/* Trap Ctrl-C */

void cleanup() {
   puts ("Caught Ctrl-C");
   quit_game ();
   _exit(0);
}

/* ignore SIGPIPE so that program does not die on write trouble */

void ignore_sigpipe () {
   struct sigaction sig;

   sig.sa_handler = SIG_IGN;
   sig.sa_flags = 0;
   sigemptyset(&sig.sa_mask);
   sigaction(SIGPIPE,&sig,NULL);
}

/* connect to server */

int make_connection (char *address, unsigned short portnum) {
   unsigned long int addr;
   struct sockaddr_in sa;
   int s;
   
   if ((addr = inet_addr(address)) < 0) 
     return (-1);
   
   sa.sin_family = AF_INET;
   sa.sin_port = htons(portnum);
   sa.sin_addr = *(struct in_addr *) address;
   
   if ((s = socket (AF_INET, SOCK_STREAM, 0)) < 0)
     return (-1);
   
   if (connect(s, &sa, sizeof(sa)) < 0)
     return (-1);

   return s;
}

/* call the shots */

int main (int ac, char *av[]) {
   struct hostent *hostinfo;
   
   if (ac < 2) {
      printf ("Usage : %s [server to connect to]", av[0]);
      exit (0);
   }
   
   if ((hostinfo = gethostbyname (av[1])) == NULL) {
      printf ("Unknown host : %s\n", av[1]);
      exit (1);
   }
   
   signal (SIGINT, cleanup);
   ignore_sigpipe();
   rl_bind_key ('\t', rl_insert);
   
   channel1 (hostinfo->h_addr_list[0]);
   return 0;
}

/* Handle all incoming replies to requests on the second channel */

void *channel2 () {
   req_t req;
   qty_t qty;
   cid_t cid;
   pr_t pr;

   while (1) { 
      sr_req2 (req);
      
      switch (req) {
       case POST_BUY_QUOT:
	 sr_cid2 (cid);
	 sr_qty2 (qty);
	 sr_pr2 (pr);
	 RECV_OK2;
	 printf ("\ncid : %d \t\t bought : %d\n", cid, qty);
	 printf ("%s", prompt_str);
	 fflush (stdout);
	 break;
	 
       case POST_SELL_QUOT:
	 sr_cid2 (cid);
	 sr_qty2 (qty);
	 sr_pr2 (pr);
	 RECV_OK2;
	 printf ("\ncid : %d \t\t sold   : %d\n", cid, qty);
	 printf ("%s",prompt_str);
	 fflush (stdout);
	 break;
	 
       default:
	 printf ("\nThis conversation is completely beyond my comprehension.\n");
	 printf ("%s",prompt_str);
	 fflush (stdout);
	 break;
      } /* switch */
   }  /* while (1) */
} /* channel2 */


void channel1 (char *address) {
   req_t req = NEW_LOGIN2;
   
   if ((server = make_connection(address, PORTNUM)) < 0) {
      perror ("make_connection");
      exit (1);
   }
   
   login_newplayer();
   
   if ((server2 = make_connection(address, PORTNUM+1)) < 0) {
      perror ("make_connection");
      exit (1);
   }
   
   sw_char (server2, req);
   sw_int (server2, passwd);
   RECV_OK2;
   
   if (ok_r == 'c') {
      int status;
      printf ("Connected and Logged in\n");

      if ((status = pthread_create (&th_b, NULL, channel2, NULL)) != 0) 
	printf ("Trouble with threads\n");
      prompt ();
   }
   else {
      printf ("Trouble\n");
      exit (2);
   }
   printf ("Trouble logging in.. bad password ?\n");
   exit (1);
}

/* issue prompt for a sort of command line */

void prompt () {
   char *line;
   
   while (1) {
      line = readline (prompt_str);
      parse_input(line);
   }
}

/* post sell or buy requests */
   
void quote (a, b, c, d) {
   req_t req1 = POST_BUY_QUOT, req2 = POST_SELL_QUOT;
   char ch;
   
   if (d == 0)     sw_req (req1);
   else sw_req (req2);

   sw_cid (a);
   sw_pr (b);
   sw_qty (c);
   
   sr_char (server, ch);
   if (ch == 'e') printf ("screwed\n");
   RECV_OK;
   return;
}

/* Parse command line */

void parse_input (char *line) {
   char *word[4], *ptr, *tmp;
   register int i=0, len=0;

   ptr = line;
   tmp = ptr;
   
   if (*ptr == '\0') return;
   
   do {
      if ((*ptr == ' ') || (*ptr == '\0')) {
	 word[i] = (char *) malloc (len);
	 strncpy (word[i], tmp, len);
	 if (*ptr == '\0') break;
	 while (*ptr == ' ') ++ptr;
	 tmp = ptr;
	 len = 0;
	 ++i;
      } else {
	 ++ptr;
	 ++len;
      }
   } while (1);

   if ((strncasecmp (word[0], "qb", 2) == 0) && (i == 3)) {	       /* quote buy */
      quote (atoi (word[1]), atoi (word[2]), atoi(word[3]), 0);
   }
   else if ((strncasecmp (word[0], "qs", 2) == 0) && (i == 3)) {   /* quote sell */
      quote (atoi (word[1]), atoi (word[2]), atoi(word[3]), 1);
   }
   else if ((strncasecmp (word[0], "i", 1) == 0) && (i == 2)) {    /* info */
      ask_comp (atoi (word[1]), atoi(word[2]));
   }
   else if (strncasecmp (word[0], "q", 1) == 0) quit_game();   /* quit */
   else if (strncasecmp (word[0], "h", 1) == 0) disp_help ();
   else {
      printf ("\nInvalid command.\n");
      disp_help ();
      return;
   }
   add_history (line);
   free (line);
   return;
}

/* Quit game */
   
void quit_game () {
   req_t req = LOGOUT;
   
   sw_req (req);
   exit (0);
}

/* New Player Login */

void login_newplayer() {
   req_t req=NEW_LOGIN;
   char *name = "Viral\0";
   short int len;
   
   len = (short) strlen (name) + 1;
   
   sw_req (req);
   sw_int (server, passwd);
   
   sw_s_int (server, len);
   sw (server, name, len);
   
   sr_int (server, plid);
   
#ifdef DEBUG
   printf ("plid %d\n", plid);
#endif
   
   assert (plid > 0);
   RECV_OK ;
}

/* Connect again, if already logged in once. */

void re_entry() {
   req_t req=RE_ENTRY;
   
   sw_req (req);
   sw_int (server, passwd);
   sw_int (server, plid);
   RECV_OK ;
}
   
void ask_comp (cid_t c1, cid_t c2) {
   req_t req1=COMP_INFO;
   req_t req2=COMP_BUY_LIST_1;
   req_t req3=COMP_SELL_LIST_1;
   

   sw_req (req1);
   sw_req (req3);
   sw_cid (c1);
   RECV_OK;
   printf ("Receiving sell list\n");
   recv_comp (c1, c1);
   
   sw_req (req1);
   sw_req (req2);
   sw_cid (c1);
   RECV_OK;
   printf ("Receiving buy list\n");
   recv_comp(c1, c1);
   
   return;
}

void recv_comp (cid_t c1, cid_t c2) {
   register int i,j;
   pr_t pr;
   qty_t qty;
   
   for (i=c1; i<c2+1; ++i) {
      printf ("cid : %d\n", i);
      printf ("pr\tqty\n");

      for (j=0; j < 4; ++j) {
	 
	 sr_pr (pr);
	 sr_qty (qty);
      
	 printf ("%d\t%d\n", pr, qty);
      }
   }
   RECV_OK;
}

void disp_help () {
   static int loop;
   static char *help[] = {
      "Experimental Client for Stock trading by Viral.\n",
      "Valid commands are :",
      "\tqb <company> <price> <quantity> :  Quote buy price.",
      "\tqs <company> <price> <quantity> :  Quote sell price.",
      "\ti <company1> <company2>         :  Get info on all companies",
      "\t                                   between company1 and company2.",
      "\tq                               :  Quit.",
      "\th                               :  Display this help.",
      NULL
   };
   
   for (loop=0; loop[help] != NULL; ++loop) {
      printf ("%s\n", help[loop]);
   }
   return;
}
